/*
 *********************************************************************

            Ejemplo de Multiplicacin rpida de un byte * 320
                  se guarda el resultado en dos bytes
                  
 **********************************************************************
*/

:BasicUpstart2(inicio)

.pc = $C000 "Ejemplo FastMul320"

// Direcciones rutinas/variables de la ROM

.const CHAROUT    = $ffd2   // salida por pantalla del caracter en A
.const CHARIN     = $ffe4   // entrada por teclado de un caracter en A
.const LINPRT     = $bdcd   // imprime el numero de 16 bits en A/X (en decimal)
.const PUTCHAR    = $ea1c   // imprime A (scancode) con color X en pos. cursor

// Variables del programa

.const numero     = $fa     // ($fa) Variable para numero a multiplicar * 320 (8 bits)
.const resultado  = $fb     // ($fb-$fc) Variable para guardar el resultado (16 bits) 
.const stringptr  = $fd     // ($fd-$fe) puntero general para cadenas de texto

msg_titulo:   .byte 147, 13
              .text "   MULTIPLICACION RAPIDA * 320"
              .byte 13,13
msg_inputnum: .text " FACTOR A MULTIPLICAR ($00-$CC)? "
              .byte 0
msg_result:   .text " * 320 = "
              .byte 0
msg_otravez:  .text " MULTIPLICAR POR OTRO FACTOR (S/N)?"
              .byte 0  
msg_error:    .byte 13,13
              .text " ERROR: MULTIPLICADOR FUERA DE RANGO" 
              .byte 13, 0

// Programa :

inicio:     // inicializa variables
            lda #0
            sta resultado
            sta numero
            
            jsr imprime_titulo          
            jsr pregunta_numero
            bcs !otravez+   
            jsr fastMUL320
            jsr imprime_resultado
            
!otravez:   lda #13
            ldx #4
!:          jsr CHAROUT
            dex
            bne !-
            
            jsr pregunta_repetir        
            bcc fin

            jmp inicio
fin:        lda #13
            jmp CHAROUT 

/* 
    Pregunta si debe repetir el programa
    devuelve la bandera C=1 en caso afirmativo, C=0 en caso negativo
*/
pregunta_repetir:
            // imprime texto mensaje de repetir (s/n)?
            lda #<msg_otravez
            sta stringptr
            lda #>msg_otravez
            sta stringptr+1
            jsr printstring
            
            // bucle de seleccion
!otra:      jsr CHARIN
            cmp #'S'
            beq !pulso_si+
            cmp #'N'
            bne !otra-      
!pulso_no:  clc
            rts
!pulso_si:  sec
            rts

/* 
    Entrada por teclado de los dos digitos hexa del factor a multiplicar * 320
    (C=0 salida sin error, C=1 salida con error)
*/
pregunta_numero:            
            jsr hexin
            tya
            asl
            asl
            asl
            asl     
            sta numero
            
            jsr hexin
            tya
            ora numero
            cmp #$cd
            bcc !+ 
            // Mensaje de error si factor >= $cd
            lda #<msg_error
            sta stringptr
            lda #>msg_error
            sta stringptr+1
            jsr printstring
            sec                 // informa de error al salir via C
!:          sta numero
            rts
/* 
    Salida por pantalla del texto de presentacion
*/          
imprime_titulo:
            lda #<msg_titulo
            sta stringptr
            lda #>msg_titulo
            sta stringptr+1
            jsr printstring
            lda #'$'
            jmp CHAROUT
/* 
    Salida por pantalla del resultado
*/
            
imprime_resultado:          

            lda #13
            jsr CHAROUT
            jsr CHAROUT
            lda #32
            jsr CHAROUT
            
            // imprime numero factor a multiplicar en decimal
            lda #0
            ldx numero
            jsr LINPRT
            
            // "* 0140 =" 
            lda #<msg_result
            sta stringptr
            lda #>msg_result
            sta stringptr+1
            jsr printstring

            // imprime en pantalla el resultado en decimal
            lda resultado+1
            ldx resultado
            jsr LINPRT
            
            // imprime resultado en hexa entre parentesis
            lda #32
            jsr CHAROUT
            lda #'('
            jsr CHAROUT
            lda #'$'
            jsr CHAROUT
            lda resultado+1
            jsr hexout
            lda resultado
            jsr hexout
            lda #')'
            jmp CHAROUT         
/*
  ***********************************************************************************************           
   Salida por pantalla de una cadena de texto terminada con ascii 0
   ($fd-$fe puntero a la cadena)
  ***********************************************************************************************
*/
printstring:
            ldy #0          
imp_texto:
            lda ($fd),y
            beq !fin+
            jsr CHAROUT
            //inc $fd
            iny
//          bne !+
//          inc $fe
//!:
            jmp imp_texto
!fin:       rts

/*
  *************************************************************************************
                       RUTINA  DE  MULTIPLICACION RAPIDA * 320
  
  ENTRADAS:
  $fa     = numero multiplicador (1 byte, rango efectivo = 0 < 204)
  
  SALIDAS:
  $fb-$fc = resultado del multiplicador * 320 (precisin 16 bits) 
  
  DETALLES:    
    Num. de ciclos          = 51 (+12 contando ciclos del JSR/RTS)
    Tamao en memoria       = 29 bytes
    Max resultado teorico   = 255 * 320 = 81600 ($013EC0) 
                            (no cabe en 16 bits, habria que adaptar resultado a 3 bytes)
    Max. resultado efectivo = 204 * 320 = 65280 ($FF00)
    
    Si se quiere optimizar para menor tamao, se pueden incluir las rotaciones
    como una parte iterada dos veces por un bucle. Pero perdemos ciclos asi...

   FECHA CREACION: 05-03-13  ,   ULT. REV.: 05-03-13                           (LOBOGRIS)
  ***************************************************************************************
*/

fastMUL320:
            // Descomponiendo, resultado de n * 320 = (256 * n)  + (64 * n)
            // ahora vamos a calcular 64*numero
            // esto se consigue poniendo numero en la parte alta del resultado para obtener numero*256
            // y desplazaremos (todos los 16 bits) de resultado a la derecha 2 veces para
            // conseguir numero*64... 
            
            lda numero
            clc                 // Primera rotacion lgica a la derecha (16 bits)
            lsr                 
            sta resultado+1     // Num. en parte alta resultado = multiplicar Num. * 256)
            lda resultado
            ror
            sta resultado       // Tras 1a rotacin, resultado = 128 * numero
        
            clc                 // 2a rotacion a la derecha (16 bits)
            lda resultado+1
            lsr                 
            sta resultado+1
            lda resultado
            ror
            sta resultado       // aqui resultado = 64 * numero
            
            // Ahora haremos resultado = (numero * 256) + (numero * 64)
            lda numero          // Recupera numero. Al ponerlo parte alta = numero * 256
            adc resultado+1     // se aade a resultado*64 (parte alta es la nica afectada)    
            sta resultado+1     // Ahora resultado = 320 * numero!
            rts


/*
  ***********************************************************************************************           
   Entrada por teclado, con acuse por pantalla de un digito hexadecimal:
   SALIDA: Se imprime por pantalla un digito hexa valido
           A contiene el ascii del digito
           Y contiene el valor binario del digito
  ***********************************************************************************************
*/
hexin:      
            lda #82         // codigo del char (no es cod ASCII, es scan code) 160 seria el espacio inverso
            ldx #14         // color del char
            jsr PUTCHAR     // dibuja un caracter en posicion de cursor
            
            jsr CHARIN      // entrada de un caracter en A
            tax             // guarda el carac. entrado en X
            cmp #$3a        // mayor o igual que ascii despues del '9'?
            bcs big
            sbc #$2f
big:        cmp #$41        // menor que cod. ASCII de 'A'?
            bcc small
            sbc #$37
small:      cmp #16
            bcs hexin       // digito > 16 ? no aceptarlo.
            tay             // guarda en Y el digito binario
            txa             // imprime el caracter valido (guardado en x)
            jsr CHAROUT
            rts
/*
  ***********************************************************************************************           
   Salida por pantalla de un byte como digitos hexadecimales:
   ENTRADA: A = byte en binario
   SALIDA: Se imprime por pantalla los dos digitos hexa que representan A
  ***********************************************************************************************
*/
 
hexout:     pha         // guarda byte completo en forma binaria
            lsr         // obtiene el primer digito hexa, dividiendo entre 16
            lsr
            lsr
            lsr
            jsr hexdig  // A representa ahora el primer digito del numero en forma binaria
            pla         // recupera el byte completo en forma binaria
            and #$0f    // obtenemos el segundo digito hexa (a traves de su nibble bajo)
            jmp hexdig  // lo imprimimos y salimos a travs de esa rutina.
            
// rutina de apoyo a hexout, impresin de 1 digito hexa 
hexdig:     cmp #$0a    // el digito es menor que 10 (digito hexa "A") ?
            bcc !+
            adc #6      // aade 7 (por que C=1) al cd. Ascii si es un digito "A"-"F"
!:          adc #48     // aade 48 (cod. del ASC"0") para obtener un digito numrico ASCII
            jmp CHAROUT // salimos a travs de CHAR OUT para ahorrarnos un RTS

            
            
            
/*          
multiplica por 2 un numero de 16 bits

en la multiplicacion de varios bytes con asl/rol 
primero se rota el byte bajo y luego teniendo en cuenta
el acarreo con rol, el byte alto. En la division el orden
de las instrucciones anlogas (lsr/ror) es al revs.
(primero se desplaza el byte alto con LSR y luego el bajo con ROR)
*/
// double16bit:
//          clc
//          lda resultado
//          asl
//          sta resultado
//          lda resultado+1
//          rol
//          sta resultado+1
//          rts
                        
// // Rutina fastmul320 - sin optimizar (usa una variable temporal de 16 bits)
//fastmul320:
//          lda #0
//          sta resultado
//          sta tmp
//          lda numero
//          sta resultado+1     // num en parte alta resultado = multiplicar * 256
//                              // ahora resultado = numero * 256
//          sta tmp+1
//          
//          clc                 // rotacion lgica a la derecha de num 16 bits
//          lda tmp+1
//          lsr                 
//          sta tmp+1
//          lda tmp
//          ror
//          sta tmp             // ahora tmp = 128 * numero
//          
//          clc                 // rotacion lgica a la derecha de num 16 bits
//          lda tmp+1
//          lsr                 
//          sta tmp+1
//          lda tmp
//          ror
//          sta tmp             // ahora tmp = 64 * numero
//          
//          // ahora sumamos las dos partes resultado = resultado + tmp
//          clc
//          lda resultado
//          adc tmp
//          sta resultado
//          lda resultado+1
//          adc tmp+1
//          sta resultado+1     // resultado = 320 * num
//          
//          rts